/*
    Tucnak - VHF contest log
    Copyright (C) 2002-2006  Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"         

struct rotar *rota,*rotb;


struct rotar *init_rotar(struct config_rotar *crot, int rotchar){
    struct rotar *rot;
    char s[20];

    if (!crot) return NULL;
    if (!crot->rot_type) return NULL;

    rot=g_new0(struct rotar, 1);
    
    switch(crot->rot_type){
        case ROT_OK1ZIA_TTYS:
            rot->sdev=sd_open_ttys(sdevlist, crot->rot_saddr, crot->rot_filename, crot->rot_timeout_ms);
            dbg("init_rotar rot=%p  saddr=%d filename='%s' timeout=%d\n", rot->sdev, crot->rot_saddr, crot->rot_filename, crot->rot_timeout_ms);
            if (!rot->sdev){
                g_free(rot);
                return NULL;
            }
            break;
#ifdef HAVE_LIBFTDI
        case ROT_OK1ZIA_FTDI:
            rot->sdev=sd_open_ftdi(sdevlist, crot->rot_saddr, crot->rot_vid, crot->rot_pid, crot->rot_serial, crot->rot_timeout_ms);
            dbg("init_rotar rot=%p  saddr=%d %04x:%04x timeout=%d\n", rot->sdev, crot->rot_saddr, crot->rot_vid, crot->rot_pid, crot->rot_timeout_ms);
            if (!rot->sdev){
                g_free(rot);
                return NULL;
            }
            break;
#endif
        case ROT_REMOTE:
            dbg("init_rotar rot=%p  saddr=%d hostname='%s' port='%d' timeout=%d\n", rot->sdev, crot->rot_saddr, crot->rot_hostname, crot->rot_port, crot->rot_timeout_ms);
            break;
        default:
            log_addf("Unsupported rotator type %d\n", crot->rot_type);
            g_free(rot);
            return NULL;
            break;

    }

    rot->desc=g_strdup(crot->rot_desc);
	rot->beamwidth=crot->rot_beamwidth;
	rot->color=0xffffffff;  /*replaced in rot_update_colors*/
/*	rot->qtf=260+40*cfg->nr;*/
//    rot->timer_id=-1;
    rot->rotchar=rotchar;
	rot->rotstr[0]=rotchar;
	rot->rotstr[1]='\0';
    rot->type = crot->rot_type;
    rot->netid = g_strdup_printf("%s:%d", crot->rot_hostname, crot->rot_port);
    rot->rem_rotstr = g_strdup(crot->rot_rem_rotstr);

    if (rot->sdev){
        rot->sdev->sconn->sprotocol = SPROT_ROTAR;
        g_snprintf(s, sizeof(s), "%d", SPROT_ROTAR);
        zwrite(rot->sdev->sconn->pipe_write, 
                zconcatesc("ADD", rot->rotstr, rot->sdev->saddr_str, s, NULL));
    }
    return rot;
}

void free_rotar(struct rotar *rot){
    g_free(rot->desc);
    g_free(rot->netid);
    g_free(rot->rem_rotstr);
    free_sd(sdevlist, rot->sdev);
    g_free(rot);
}

int init_rotars(){
    struct rotar *rot;
    struct config_rotar *cfgrot;
    int i;
    char rotchar='A';

    rotars=g_ptr_array_new();
    for (i=0;i<cfg->crotars->len;i++){
        cfgrot=(struct config_rotar*)g_ptr_array_index(cfg->crotars, i);
        if (!cfgrot) continue;
        rot=init_rotar(cfgrot, rotchar++);
        if (!rot) continue;
        g_ptr_array_add(rotars, rot); 
    }
#ifdef HAVE_SDL
    rot_update_colors();
#endif
    return 0;
}

int free_rotars(void){
    int i;
    struct rotar *rot;
    for (i=0;i<rotars->len;i++){
        rot=g_ptr_array_index(rotars, i);
        free_rotar(rot);
    }
    g_ptr_array_free(rotars,1);
    rotars=NULL;
    return 0;
}

#ifdef HAVE_SDL
int rot_update_colors(){
	int i;
	struct rotar *rot;
	
    if (!sdl || !rotars) return;
    for (i=0;i<rotars->len;i++){
        rot=g_ptr_array_index(rotars, i);
	
		switch(i%4){
			case 0:
				rot->color=makecol(200,0,0);
				break;
			case 1:
				rot->color=makecol(200,0,200);
				break;
			case 2:
				rot->color=makecol(0,200,0);
				break;
			case 3:
				rot->color=makecol(100,200,100);
				break;
		}
	}
	return 0;
}
#endif

int rot_seek(struct rotar *rot, int uhel){
    char s[10], *c;

    if (!rot) return 0;

    switch (rot->type){
        case ROT_REMOTE:
            c = g_strdup_printf("RS %s;%s;%d;%s\n", 
                    rot->netid, rot->rem_rotstr, uhel, "");
            rel_write_all(c);
            g_free(c);
            
            break;
        default:
            g_snprintf(s, sizeof(s), "%d", uhel);
            zwrite(rot->sdev->sconn->pipe_write, 
                zconcatesc("AZIM", rot->sdev->saddr_str, s, NULL));
            break;
    }
    return 0;
}

struct config_rotar *get_config_rotar_by_number(GPtrArray *crots, int nr){
    struct config_rotar *crot;
    int i;

    for (i=0; i<cfg->crotars->len; i++){
        crot = (struct config_rotar *)g_ptr_array_index(crots, i);
        if (crot->nr==nr) return crot;
    }
    return NULL;   
}

struct rotar *get_rotar(int nr){
    if (!rotars) return NULL;
    if (nr>=rotars->len) return NULL;
    return (struct rotar*) g_ptr_array_index(rotars, nr);
}

struct rotar *get_rotar_by_rotstr(char *rotstr){
    int i;
    struct rotar *rot;

    if (!rotars) return NULL;
    for (i=0;i<rotars->len;i++){
        rot=(struct rotar *)g_ptr_array_index(rotars, i);
        if (rot->rotchar==rotstr[0]) return rot;
    }
    return NULL;
}

struct rotar *get_rotar_by_netid_rotstr(char *netid, char *rotstr){
    int i;
    struct rotar *rot;

    if (!rotars) return NULL;
    for (i=0;i<rotars->len;i++){
        rot=(struct rotar *)g_ptr_array_index(rotars, i);
        dbg("  rot->type=%d rotchar=%s netid=%s\n", rot->type, rot->rem_rotstr, rot->netid);
        if (rot->type != ROT_REMOTE) continue;
        if (strcmp(rot->netid, netid)!=0) continue;
        if (strcmp(rot->rem_rotstr, rotstr)!=0) continue;
        return rot;
    }
    return NULL;
}


void rotar_read_handler(char *line){
    struct zstring *zs, *zs2;
    char *cmd, *rotstr, *c;
	struct rotar *rot;
    
    zs = zstrdup(line);
                          
    ztokenize(zs, 1);
	rotstr = ztokenize(zs, 0);

	rot = get_rotar_by_rotstr(rotstr);
	if (!rot){
		log_addf("Unknown rotar '%s'\n", rotstr); 
		dbg("Unknown rotar '%s'\n", rotstr); 
		goto x;
	}
	cmd = ztokenize(zs, 0);
	
//    dbg("rotar_read_handler   rcvd: '%s'  cmd='%s' rotstr='%s'\n", line, cmd, rotstr);
    if (strcmp(cmd, "!")==0){  /* error */
        log_addf("Rotar error: %s", ztokenize(zs, 0));
    }
    if (strcasecmp(cmd, "QTF")==0){
        char *qtf = ztokenize(zs, 0);
        rot->qtf=atoi(qtf);
        
        zs2 = zconcatesc(net->myid, rotstr, qtf, "0", NULL);
        c = g_strconcat("RU ", zs2->str, "\n", NULL);
        rel_write_all(c);
        g_free(c);
        zfree(zs2);
        
    }
x:;	
    zfree(zs);
    redraw_later();
#ifdef HAVE_SDL
    if (gses && gses->ontop->type == SWT_MAP) 
//        sw_map_redraw_rect(gses->ontop, NULL);
        sw_map_update_rotar(gses->ontop);
#endif

}


void rotar_remote_update(char *c){
    char *netid, *rotstr, *qtf, *elev;
    struct zstring *zs;
    struct rotar *rot;

    dbg("rotar_remote_update('%s')\n", c);
    zs = zstrdup(c);

    netid = ztokenize(zs, 1);
    if (!netid) goto x;
    rotstr = ztokenize(zs, 0);
    if (!rotstr) goto x;
    qtf = ztokenize(zs, 0);
    if (!qtf) goto x;
    elev = ztokenize(zs, 0);
    if (!elev) goto x;

    rot = get_rotar_by_netid_rotstr(netid, rotstr);
    dbg("rotar_remote_update id='%s' rotstr='%s'   rot=%p\n", netid, rotstr, rot);

    if (!rot) goto x;

    rot->qtf = atoi(qtf);
    rot->elev = atoi(elev);
    redraw_later();
x:;
    zfree(zs);
    
}


/****************************** rotar dialog ***************************/

static char rotar_qtf_str[EQSO_LEN], rotar_elev_str[EQSO_LEN];
static char rotar_desc1[MAX_STR_LEN], rotar_desc2[MAX_STR_LEN];

void refresh_rotar(void *xxx){
    if (strlen(rotar_qtf_str)>0){
        rot_seek(rotar, atoi(rotar_qtf_str));
    }
    /*dbg("refresh_rotar\n");*/
}

char *rotar_msg[] = {
    CTEXT(T_AZIMUTH),
    CTEXT(T_ELEVATION),
};

void rotar_fn(struct dialog_data *dlg)
{
    struct terminal *term = dlg->win->term;
    int max = 0, min = 0;
    int w, rw;
    int y = -1;

    max_text_width(term, rotar_desc1, &max);
    min_text_width(term, rotar_desc1, &min);
    max_text_width(term, rotar_desc2, &max);
    min_text_width(term, rotar_desc2, &min);
    max_group_width(term, rotar_msg + 0, dlg->items + 0, 1, &max);
    min_group_width(term, rotar_msg + 0, dlg->items + 0, 1, &min);
    max_group_width(term, rotar_msg + 1, dlg->items + 1, 1, &max);
    min_group_width(term, rotar_msg + 1, dlg->items + 1, 1, &min);
    
    max_buttons_width(term, dlg->items + 2, 2, &max);
    min_buttons_width(term, dlg->items + 2, 2, &min);
    
    w = dlg->win->term->x * 9 / 10 - 2 * DIALOG_LB;
    if (w > max) w = max;
    if (w < min) w = min;
    if (w > dlg->win->term->x - 2 * DIALOG_LB - 8 ) w = dlg->win->term->x - 2 * DIALOG_LB - 8;
    if (w < 1) w = 1;
    
    rw = 0;
    y ++;
    dlg_format_text (NULL, term, rotar_desc1, dlg->x+6, &y, w, &rw, COLOR_DIALOG_TEXT, AL_LEFT);
    dlg_format_text (NULL, term, rotar_desc2, dlg->x+6, &y, w, &rw, COLOR_DIALOG_TEXT, AL_LEFT);
    dlg_format_group(NULL, term, rotar_msg + 0, dlg->items + 0, 1, 0, &y, w, &rw);
    dlg_format_group(NULL, term, rotar_msg + 1, dlg->items + 1, 1, 0, &y, w, &rw);
    y++;
    dlg_format_buttons(NULL, term, dlg->items + 2, 2, 0, &y, w, &rw, AL_LEFT);
    
    
    w = rw;
    dlg->xw = w + 2 * DIALOG_LB;
    dlg->yw = y + 2 * DIALOG_TB;

    
    center_dlg(dlg);
    draw_dlg(dlg);
    y = dlg->y + DIALOG_TB;
    y++;
    dlg_format_text (term, term, rotar_desc1, dlg->x+6, &y, w, &rw, COLOR_DIALOG_TEXT, AL_LEFT);
    dlg_format_text (term, term, rotar_desc2, dlg->x+6, &y, w, &rw, COLOR_DIALOG_TEXT, AL_LEFT);
    dlg_format_group(term, term, rotar_msg + 0, dlg->items + 0, 1, dlg->x + DIALOG_LB, &y, w, AL_LEFT);
    dlg_format_group(term, term, rotar_msg + 1, dlg->items + 1, 1, dlg->x + DIALOG_LB, &y, w, AL_LEFT);
    y++;
    dlg_format_buttons(term, term, dlg->items + 2, 2, dlg->x + DIALOG_LB, &y, w, NULL, AL_LEFT);
    
}   

void rotar_set_desc(void){
    switch(rotar->type){
        case ROT_REMOTE:
            g_snprintf(rotar_desc1, MAX_STR_LEN, "Rotr %c (%s %s)", rotar->rotchar, rotar->netid, rotar->rem_rotstr);
            break;
        default:
            g_snprintf(rotar_desc1, MAX_STR_LEN, "Rotr %c (%u)", rotar->rotchar, (unsigned char)rotar->sdev->saddr);
            break;
    }
    g_snprintf(rotar_desc2, MAX_STR_LEN, "%s", rotar->desc);
}

int rotar_func(struct dialog_data *data, struct event *ev){
    char rotchar;
      /*  dbg("mouse ev=%d x=%d y=%d b=%d\n", ev->ev, ev->x, ev->y, ev->b); */
    
    switch(ev->ev){
        case EV_KBD:
            if (ev->y & KBD_ALT) break;
            if (ev->y & KBD_CTRL) break;
            rotchar=upcase(ev->x);
            if (rotchar<'A' || rotchar>'Z') {
                return -1;
            }
            if (rotchar<'A' || rotchar-'A'>=rotars->len) {
                return EVENT_PROCESSED;
            }
            /*dbg("rotar %c\n", rotchar);*/

            rotar=get_rotar(rotchar-'A');
            rotar_set_desc();
            resize_terminal(CBA0);
            return EVENT_PROCESSED;
        
    }
    return -1; /* !EVENT_PROCESSED */
}
        
        

void menu_rotar(union cba_t cba){

    struct dialog *d;
    int i;

    if (!rotar) rotar=get_rotar(0);
    if (!rotar) return;
    
    rotar_set_desc();
    /*g_snprintf(rotar_qtf_str,  EQSO_LEN, "%d", rotar->qtf);
    g_snprintf(rotar_elev_str, EQSO_LEN, "%d", rotar->elev);*/
    strcpy(rotar_qtf_str, "");
    strcpy(rotar_elev_str, "");

    
    if (!(d = mem_alloc(sizeof(struct dialog) + 55 * sizeof(struct dialog_item)))) return;
    memset(d, 0, sizeof(struct dialog) + 55 * sizeof(struct dialog_item));
    d->title = "Rotator";
    d->fn = rotar_fn;
    d->refresh = (void (*)(void *))refresh_rotar;
    d->refresh_data = (void *)rotar;
    d->handle_event = rotar_func;
    
    d->items[i=0].type = D_FIELD; /* 0 */
    d->items[i].dlen = EQSO_LEN;
    d->items[i].data = rotar_qtf_str;
    
    d->items[++i].type = D_FIELD; 
    d->items[i].dlen = EQSO_LEN;
    d->items[i].data = rotar_elev_str;
    
   

    d->items[++i].type = D_BUTTON; /* 2 */
    d->items[i].gid = B_ENTER;
    d->items[i].fn = ok_dialog;
    d->items[i].text = VTEXT(T_OK);
    
    d->items[++i].type = D_BUTTON;
    d->items[i].gid = B_ESC;
    d->items[i].fn = cancel_dialog;
    d->items[i].text = VTEXT(T_CANCEL);
    
    
    d->items[++i].type = D_END;
    do_dialog(d, getml(d, NULL));
                               
}

