/*
    Tucnak - VHF contest log
    Copyright (C) 2002-2006  Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"
    

#define STARTB 4
#define STOPB 1


struct sslave{
    struct sslave *next;
    struct sslave *prev;

    unsigned char saddr;
    enum sprotocol sprotocol;
    char rotstr[5];

    /* for SPROT_ROTAR */
    int oldqtf;
};


struct sdevlist *init_sdevlist(){
    struct sdevlist *sdevlist;

    sdevlist=g_new0(struct sdevlist,1);
    sdevlist->sdevs=g_ptr_array_new();
    sdevlist->sconns=g_ptr_array_new();
    return sdevlist;
}

int free_sdevlist(struct sdevlist *sdevlist){
    int i;
    for (i=0;i<sdevlist->sdevs->len;i++){
        struct sdev *sd;
        sd=(struct sdev *)g_ptr_array_index(sdevlist->sdevs, i);
        free_sd(sdevlist, sd);
    }
    g_free(sdevlist);
    return 0;
}


struct sdev *sd_open_ttys(struct sdevlist *sdevlist, char saddr, char *filename, int timeout_ms){
    int i;
    struct sconn *sc;
	struct sdev *sd;
    
/*    dbg("sd_open_ttys(%s)\n", filename);*/
    
    for (i=0;i<sdevlist->sconns->len;i++){
        sc=(struct sconn *)g_ptr_array_index(sdevlist->sconns, i);
        if (sc->type!=CT_TTYS) continue;
        if (strcmp(sc->ttys_filename, filename)!=0) continue;
        /*dbg("found sc=%p\n", sc);*/
    	sc->refcnt++;
        goto found;
    }
    /* sconn not found */
    sc=g_new0(struct sconn,1);
    dbg("new sc %p\n",sc);
    sc->refcnt++;
    g_ptr_array_add(sdevlist->sconns, sc);
//    init_list(sc->sjobq);
    sc->type=CT_TTYS;
    sc->ttys_filename=g_strdup(filename);
    sc->pipe_write = -1;
    sc->timeout_ms = timeout_ms;
    sc->sprotocol = SPROT_NONE;
    sc->open  = sconn_file_open;
    sc->close = sconn_file_close;
    sc->read  = sconn_file_read;
    sc->write = sconn_file_write;

    sc_open_common(sc);
    
found:    
    sd=g_new0(struct sdev,1);
    g_ptr_array_add(sdevlist->sdevs, sd);

    sd->sconn=sc;
    sd->saddr=saddr;
    g_snprintf(sd->saddr_str, 5, "%d", (unsigned char)saddr);
    /*dbg("sd=%p refcnt=%d\n",sd,sc->refcnt);*/
    return sd;
}

    
#ifdef HAVE_LIBFTDI
struct sdev *sd_open_ftdi(struct sdevlist *sdevlist, char saddr, int vid, int pid, char *serial, int timeout_ms){
    int i;
    struct sconn *sc;
	struct sdev *sd;
    
/*    dbg("sd_open_ttys(%s)\n", filename);*/
    
    for (i=0;i<sdevlist->sconns->len;i++){
        sc=(struct sconn *)g_ptr_array_index(sdevlist->sconns, i);
        if (sc->type!=CT_FTDI) continue;
        if (serial && *serial)
            if (strcmp(sc->ftdi_serial, serial)!=0) continue;
        /*dbg("found sc=%p\n", sc);*/
    	sc->refcnt++;
        goto found;
    }
    /* sconn not found */
    sc=g_new0(struct sconn,1);
    dbg("new sc %p\n",sc);
    sc->refcnt++;
    g_ptr_array_add(sdevlist->sconns, sc);
//    init_list(sc->sjobq);
    sc->type=CT_FTDI;
    sc->pipe_write = -1;
    sc->timeout_ms = timeout_ms;
    sc->ftdi_vid = vid;
    sc->ftdi_pid = pid;
    sc->ftdi_serial = g_strdup(serial?serial:"");
    sc->sprotocol = SPROT_ROTAR;
    sc->open  = sconn_ftdi_open;
    sc->close = sconn_ftdi_close;
    sc->read  = sconn_ftdi_read;
    sc->write = sconn_ftdi_write;

    sc_open_common(sc);
    
found:    
    sd=g_new0(struct sdev,1);
    g_ptr_array_add(sdevlist->sdevs, sd);

    sd->sconn=sc;
    sd->saddr=saddr;
    g_snprintf(sd->saddr_str, 5, "%d", (unsigned char)saddr);
    /*dbg("sd=%p refcnt=%d\n",sd,sc->refcnt);*/
    return sd;
}
#endif // HAVE_LIBFTDI
    
    
void sc_open_common(struct sconn *sc){
    int i,fds[2];

    if (pipe(fds)) internal("Can't create pipe");
    sc->pipe_write=fds[1];
    sc->pid=fork();
    if (!sc->pid){
        close(0);
        dup(fds[0]);
        close(1);
        dup(tpipe->threadpipe_write);
        for (i=3; i<1024; i++) close(i);
        sc_main(sc);
        exit(0);
    }
    if (sc->pid<0) {
        internal("Can't fork for sconn\n");
    }
}

struct sdev *sd_open_udp(char *hostname, int udpport){
    return NULL;    
}

struct sdev *sd_open_tcp(char *hostname, int tcpport){
    
    return NULL;
}

int free_sd(struct sdevlist *sdevlist, struct sdev *sd){
    struct sconn *sc;
    
/*    dbg(" ------\n");*/
    dbg("free_sd(%p)\n", sd);

    if (!sd) return -1;

    sc=sd->sconn;
   /* dbg("  refcnt=%d\n", sc->refcnt);*/
    sc->refcnt--;
    if (!sc->refcnt){
        int status;
        
        /*dbg("  freeing sc %p\n", sc);*/
        if (sc->pid) kill(sc->pid, SIGINT);
        if (sc->pipe_write>=0) close(sc->pipe_write);
        waitpid(sc->pid, &status, 0);
        sc->pid=0;
        CONDGFREE(sc->ttys_filename);
        CONDGFREE(sc->ip_hostname);
#ifdef HAVE_LIBFTDI
        CONDGFREE(sc->ftdi_serial);
#endif
        g_ptr_array_remove(sdevlist->sconns, sc);
        g_free(sc);
    }else{
        /*dbg("  sc %p already active\n", sc);*/
    }
    g_ptr_array_remove(sdevlist->sdevs, sd);
    g_free(sd);
    return (0);
}

static char sd_chk(unsigned char *s, int len){
	unsigned char chk, *c;
    chk=0;
	for (c=s;len;c++,len--){
		chk^=*c;
	}
	return chk;
}


int sd_prot(struct sconn *sconn, char saddr, char fce, char *data, int *len, int timeout){
	unsigned char rawdata[550];
	int rawlen,written,rawi;
    int ret, i;


    /* clearing queue, filedescriptor is non-blocking */
    while (sconn->read(sconn, rawdata, sizeof(rawdata)-1, 100)>0){};
    
	rawlen=0;
	memset(rawdata, 0xff, STARTB); rawlen+=STARTB;
	rawdata[rawlen++]=0xc5;
	rawdata[rawlen++]=fce&0x7f;
	rawdata[rawlen++]=saddr;
	rawdata[rawlen++]=(unsigned char)*len;
    memcpy(rawdata+rawlen, data, *len); rawlen+=*len;
	rawdata[rawlen]=sd_chk(rawdata+STARTB, rawlen-STARTB);
    rawlen++;
	memset(rawdata+rawlen, 0xff, STOPB); rawlen+=STOPB;
	
	written=sconn->write(sconn, rawdata, rawlen);
   /* dbg("sd_send: written=%d\n", written);*/
    if (written<0) return written;

    rawi=0;
    while(1){
        if (rawi>=sizeof(rawdata)-1) return 20;

        
        ret = sconn->read(sconn, rawdata+rawi, sizeof(rawdata)-rawi, sconn->timeout_ms);
//        dbg("read=%d\n", ret);
        if (ret<0) return -3;
        if (ret==0) return -4;
        rawi += ret;
#if 0        
        {
            int j;
            dbg("read=");
            for (j=0; j<rawi; j++) dbg("%02x ", (unsigned char)rawdata[j]);
            dbg("\n");
        }
#endif        
        
        for (i=0; i<rawi; i++){
            if (rawdata[i]!=0xc5) continue;
#if 0            
            {
                int j;
                dbg("c5 at %d\n", i);
                for (j=0; j<rawi; j++) dbg("%02x ", (unsigned char)rawdata[j]);
                dbg("\n");
            }
#endif            
           // dbg("i+5>rawi %d+5>%d=%d\n", i, rawi, i+5>rawi);
            if (i+5>rawi) goto nextloop;
           // dbg("i+5+rawdata[i+3]>rawi %d+5+%d>%d\n", i, (unsigned char)rawdata[i+3], rawi);
            if (i+5+(unsigned char)rawdata[i+3]>rawi) goto nextloop;
           // dbg("b\n");
            if (sd_chk(rawdata+i, 5+rawdata[i+3]) != 0) return 11;
            if (rawdata[i+1]==0) return 17;
            if (rawdata[i+1]==0x80) return 14;
            if ((rawdata[i+1] & 0x80)==0) continue;
//            dbg("c\n");
            if (rawdata[i+2]!=(unsigned char)saddr) return 16;
//            dbg("d\n");
            if (rawdata[i+1]!=(fce|0x80)) return 16;

            *len = rawdata[i+3];
            memcpy(data, rawdata+i+4, *len);
/*            {
                int j;
                dbg("OK data=\n");
                for (j=0; j<*len; j++) dbg("%02x ", (unsigned char)data[j]);
                dbg("\n");
            }*/
            return 0;
        }
        //dbg("neni C5 0..%d\n", rawi);
nextloop:;        
    }
    

    
    return 0;
}



char *sd_err(int err){
    switch(err){
        case 0:
            return "OK";
        case 11:
            return "Bad checksum";
        case 13:
            return "Timeout";
        case 14:
            return "Error";
        case 16:
            return "Bad response";
        case 17:
            return "Unknown function";
        default:
            return "Unknown error code";
    }

}

int sconn_file_open(struct sconn *sconn, int verbose){
	struct termios tio;
    char s[256], errbuf[256];
    int i, ret;
	
    ret = fhs_lock(sconn->ttys_filename, 1);
    if (ret){
        fhs_error(s, sizeof(s), ret, sconn->ttys_filename);
        zwrite(1, zconcatesc("SC", "!", s, NULL));
        return ret;
    }

	sconn->fd = open(sconn->ttys_filename, O_RDWR | O_SYNC | O_NONBLOCK | O_NOCTTY);
    if (sconn->fd<0){
        g_snprintf(s, sizeof(s), "Can't open device %s", sconn->ttys_filename);
        zwrite(1, zconcatesc("SC", "!", s, NULL));
        fhs_unlock(sconn->ttys_filename);

		return -3;
    }
    
    if (fcntl(sconn->fd,F_SETFL,O_NONBLOCK)) {
		g_snprintf(s, sizeof(s), "Can't set O_NONBLOCK on %s: %s",sconn->ttys_filename, strerror_r(errno, errbuf, sizeof(errbuf)) );
		zwrite(1, zconcatesc("SC", "!", s, NULL));
        fhs_unlock(sconn->ttys_filename);
        return -4;
    }

#if 0
    // old buggy code
	tcgetattr(sconn->fd,&tio);
	tio.c_cflag=B9600|CS8|CLOCAL|CREAD|PARENB;/*|PARODD;*/
	tio.c_iflag=INPCK;
	tio.c_lflag=0;
	tio.c_oflag=0;
	tio.c_cc[VMIN]=1;
	tio.c_cc[VTIME]=5;
	tcsetattr(sconn->fd,TCSANOW,&tio);
#else
	tcgetattr(sconn->fd,&tio);
    cfmakeraw(&tio);
    cfsetispeed(&tio, B9600);
    cfsetospeed(&tio, B9600);
    tio.c_cflag |= (CLOCAL | CREAD);
    tio.c_cflag &= ~CSIZE;
    tio.c_cflag |= CS8;
    tio.c_cflag &= ~CSTOPB;
    tio.c_cflag |= PARENB;
    tio.c_cflag &= ~PARODD;
    tio.c_cflag &= ~CRTSCTS;
    tio.c_iflag &= ~IXON;
    tio.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);
    tio.c_oflag &= ~OPOST;
    tcflush(sconn->fd, TCIFLUSH);
    tcsetattr(sconn->fd, TCSANOW, &tio);
#endif

    i = TIOCM_RTS; 
    ioctl (sconn->fd, TIOCMBIS, &i); /* set */
    i = TIOCM_DTR; 
    ioctl (sconn->fd, TIOCMBIS, &i); /* set */

/* 
 * normally 0
 * 1 for debugging with atisp
 *
 */

#if 1    
    i = TIOCM_RTS; /* clear PSEN*/
    ioctl (sconn->fd, TIOCMBIC, &i); /* clear = log.1*/
    //usleep(100000);
    i = TIOCM_DTR; /* clear reset*/
    ioctl (sconn->fd, TIOCMBIC, &i); /* clear */
#endif
    return 0;
}

int sconn_file_close(struct sconn *sconn){
    
    //dbg("sconn_file_close %s\n",sconn->ttys_filename); 
    close(sconn->fd);
    fhs_unlock(sconn->ttys_filename);
    return -1;
}

int sconn_file_read(struct sconn *sconn, void *data, size_t len, int timeout_ms){
    int ret;
    fd_set fdr;
    struct timeval tv;
        
    FD_ZERO(&fdr);
    FD_SET(sconn->fd, &fdr);
    tv.tv_usec = timeout_ms * 1000;
    tv.tv_sec = 0;
    
    ret = select(sconn->fd+1, &fdr, NULL, NULL, &tv);
//        dbg("select=%d\n", ret);
    if (ret<0) return -1;
    if (ret==0) return 13;
    if (!FD_ISSET(sconn->fd, &fdr)) return -2;

    return read(sconn->fd, data, len);
}

int sconn_file_write(struct sconn *sconn, void *data, size_t len){
    return write(sconn->fd, data, len);
}


#ifdef HAVE_LIBFTDI

int sconn_ftdi_open(struct sconn *sconn, int verbose){
    int ret;
    
    sconn->ftdi = ftdi_new();
    if (!sconn->ftdi){
        if (verbose) log_addf("Can't init ftdi library for sdev");
        return -1;
    }
    if (sconn->ftdi_serial && *sconn->ftdi_serial){
        ret = ftdi_usb_open_desc(sconn->ftdi, sconn->ftdi_vid, sconn->ftdi_pid, NULL, sconn->ftdi_serial);
        dbg("ftdi_usb_open(%04x:%04x, '%s')=%d\n", sconn->ftdi_vid, sconn->ftdi_pid, sconn->ftdi_serial, ret);
    }else{
        ret = ftdi_usb_open(sconn->ftdi, sconn->ftdi_vid, sconn->ftdi_pid);
        dbg("ftdi_usb_open(%04x:%04x)=%d\n", sconn->ftdi_vid, sconn->ftdi_pid, ret);
    }
    if (ret){
        if (verbose) {
            log_addf("Can't open ftdi device %04x:%04x, error=%d %s", sconn->ftdi_vid, sconn->ftdi_pid, ret, sconn->ftdi->error_str);
            if (ret==-8) log_addf("Maybe try to run as root: \"adduser %s dialout\" and relogin", getenv("USER"));
        }
        return -2;

    }
    ret = ftdi_set_baudrate(sconn->ftdi, 9600);
    if (ret){
        if (verbose) log_addf("Can't set baudrate for sconn, error=%d %s", ret, sconn->ftdi->error_str);
        return -3;
    }
    ret = ftdi_set_line_property(sconn->ftdi, BITS_8, STOP_BIT_1, EVEN);
    if (ret){
        if (verbose) log_addf("Can't set line properties for sconn, error=%d %s", ret, sconn->ftdi->error_str);
        return -4;
    }
    return 0;
}

int sconn_ftdi_close(struct sconn *sconn){
    if (sconn->ftdi!=NULL){
        ftdi_free(sconn->ftdi);
        sconn->ftdi=NULL;
    }
    return 0;
}

int sconn_ftdi_read(struct sconn *sconn, void *data, size_t len, int timeout_ms){
    int ret;
    struct timeval start, stop;
    int usec, sec;
    
    
    gettimeofday(&start, NULL);
    while(1){
        ret = ftdi_read_data(sconn->ftdi, data, len);
        if (ret) break;
        gettimeofday(&stop, NULL);
        usec = stop.tv_usec - start.tv_usec;
        sec = stop.tv_sec - start.tv_sec;
        if (usec<0) usec+=1000000,sec--;
        if (sec>timeout_ms/1000) break;
        if (usec/1000>timeout_ms) break;
        usleep(1000);
    }
    
//    dbg("sconn_ftdi_read(len=%d)=%d\n", len, ret);
    return ret;
}

int sconn_ftdi_write(struct sconn *sconn, void *data, size_t len){
    int ret;
    ret = ftdi_write_data(sconn->ftdi, data, len);
//    dbg("sconn_ftdi_write(len=%d)=%d\n", len, ret);
    return ret;
}

#endif // HAVE_LIBFTDI

static int konec = 1;

static void sigint(int no){
    konec = 1;    
}

int sc_main(struct sconn *sconn){
	char s[256], data[256];
	int len, ret;
    struct sslave *sptr; 
    
    signal(SIGINT, sigint);
    signal(SIGSEGV, SIG_DFL);
	
    dbg("sc_main pid=%d\n", getpid());
    
    struct sslave slaves;
        
    init_list(slaves);
    slaves.sprotocol = SPROT_NONE;
    
    ret = sconn->open(sconn, 1);
    if (ret) return ret;
    konec = 0;
    
    if (fcntl(0, F_SETFL, O_NONBLOCK)){
        printf("SC;i;Can't set O_NONBLOCK\n");
        return -5;
    }

    sptr = &slaves;
 //   dbg("slaves=%p\n", &slaves);
    while(!konec){
        struct zstring *zs;
        char *cmd;
        int qtf,slen;
        struct sslave *slave;

        
        s[sizeof(s)-1]='\0';
        if (fgets(s, sizeof(s)-1, stdin) != NULL){
            slen=strlen(s);
            if (slen>0 && s[slen-1]=='\n') s[slen-1]='\0';
//            dbg("sc_main s='%s'\n", s);
            zs = zstrdup(s);
            cmd = ztokenize(zs, 1);

            if (strcasecmp(cmd, "ADD")==0){
                slave = g_new0(struct sslave, 1);
                add_to_list(slaves, slave);
                safe_strncpy(slave->rotstr, ztokenize(zs, 0), 5);
                slave->saddr=atoi(ztokenize(zs, 0));
                slave->sprotocol=atoi(ztokenize(zs, 0));
                slave->oldqtf=-1;
                dbg("add slave=%p\n", slave);
            }
            
            if (strcasecmp(cmd, "DEL")==0){
                unsigned char saddr2;

                saddr2=atoi(ztokenize(zs, 0));
                foreach(slave, slaves){
                    if (slave->saddr!=saddr2) continue;
                    del_from_list(slave);
                    g_free(slave);
                    break;
                }
                sptr = &slaves; // nebudeme zkoumat, co se zmenilo, vezmeme to od zacatku
            }
            if (strcasecmp(cmd, "AZIM")==0){
                unsigned int azim;
                unsigned char saddr;

                saddr = atoi(ztokenize(zs, 0));
                azim = atoi(ztokenize(zs, 0));
                data[0]=azim&0xff;
                data[1]=(azim>>8)&0xff;
                len=2;
                sd_prot(sconn, saddr, 65, data, &len, 400);
            }
        }
#if 0
        
        
        sprintf(s, "%d", ret++);
        if (ret==360) ret=0;
        zwrite(1, zconcatesc("ROT", "A", "QTF", s, NULL));
        usleep(10000);
        continue;
#endif
        
        
//        dbg("sptr=%p  saddr=%d sprotocol=%d\n", sptr, sptr->saddr, sptr->sprotocol);
        switch(sptr->sprotocol){
            case SPROT_ROTAR:
                len=0;
                ret=sd_prot(sconn, sptr->saddr, 64, data, &len, sconn->timeout_ms);
//                dbg("sd_prot(64)=%d\n", ret);
                
                if (ret!=0){
                    break;
                }
                qtf=(unsigned char)data[0]+256*(unsigned char)data[1];
                if (qtf==sptr->oldqtf) break;
                
                sptr->oldqtf=qtf;
                g_snprintf(s, sizeof(s), "%d", qtf);
                zwrite(1, zconcatesc("ROT", sptr->rotstr, "QTF", s, NULL));
                break;
            default:
                break;
        }
        sptr=sptr->next;
        if (sptr==&slaves) sptr=sptr->next;
    }
    dbg("sc_main exiting\n");
    sconn->close(sconn);
    return 0;
}

