/*
    Tucnak - VHF contest log
    Copyright (C) 2002-2006  Ladislav Vaiz <ok1zia@nagano.cz>

    This program is free software; you can redistribute it and/or                                                        
    modify it under the terms of the GNU General Public License                                                          
    version 2 as published by the Free Software Foundation.

*/

#include "header.h"

#ifdef HAVE_HAMLIB

struct trig *trig;
GIndexArray *riglist = NULL;


struct trig *init_trig(){
	struct trig *trig;
    int i, fds[0], ret;
    char s[256];

  //  dbg("init_trig()\n");
    rig_set_debug(RIG_DEBUG_BUG);
    if (!riglist){
        riglist = g_index_array_new();
        rig_load_all_backends();
        rig_list_foreach(trig_save_riglist, riglist);
        g_index_array_qsort(riglist, trig_compare);
#if 0
        for (i=0; i<riglist->len; i++){
            struct rig_caps *caps = (struct rig_caps *)g_index_array_index(riglist, i);
            dbg("%4d %4d %-20s %s\n", i, caps->rig_model, caps->mfg_name, caps->model_name);
        }
        dbg("X\n");
#endif
    }
    

    
    if (!cfg->rig_model) return NULL;
    if (!cfg->rig_filename || !*cfg->rig_filename) return NULL;

	trig = g_new0(struct trig, 1);
    trig->pipe_write = -1;
    trig->qrg = 0.0;
   

    ret = fhs_lock(cfg->rig_filename, 1);
    if (ret){
        fhs_error(s, sizeof(s), ret, cfg->rig_filename);
        log_addf("Rig error: %s", s);
        free_trig(trig);
        return NULL;
    }
    trig->locked = 1;
    
    if (pipe(fds)) internal("Can't create pipe");
    trig->pipe_write=fds[1];
    trig->pid=fork();
    if (!trig->pid){
        close(0);
        dup(fds[0]);
        close(1);
        dup(tpipe->threadpipe_write);
        for (i=3; i<1024; i++) close(i);
        trig_main((rig_model_t)cfg->rig_model, cfg->rig_filename);
        exit(0);
    }
    if (trig->pid<0) {
        close(trig->pipe_write);
        log_addf("Can't fork for trig\n");
        g_free(trig);
        return NULL;
    }
    dbg("trig->pid=%d\n", trig->pid);
	return trig;
}

void free_trig(struct trig *trig){
    int status;
    
    if (riglist){
        g_index_array_free(riglist, TRUE);
        riglist = NULL;
    }
    
    
    if (!trig) return;

    if (!trig->pid) return;
    ST_START;
    if (trig->pid) kill(trig->pid, SIGINT);
    waitpid(trig->pid, &status, 0);
    ST_STOP;
  /*  dbg("done\n");*/
    trig->pid=0;
	if (trig->pipe_write>=0) close(trig->pipe_write);
    if (trig->locked) fhs_unlock(cfg->rig_filename);
	g_free(trig);
}

void trig_read_handler(struct trig *trig, char *s){
    struct zstring *zs;
    char *cmd;
    
    zs = zstrdup(s);
    
    cmd = ztokenize(zs, 1);
//     dbg("trig_read_handler   rcvd: '%s'  cmd='%s'\n", s, cmd);
    if (strcmp(cmd, "!")==0){  /* error */
        log_addf("Rig error: %s", ztokenize(zs, 0));
    }
    if (strcasecmp(cmd, "qrg")==0){
        trig->qrg = atof(ztokenize(zs,0)) + get_rig_lo(aband);
    }
    zfree(zs);
    redraw_later();
}


void trig_set_mode(struct trig *trig, rmode_t mode){
    char s[40];

    if (!trig) return;
    g_snprintf(s, sizeof(s), "MODE;%d;\n", (int)mode);
    write(trig->pipe_write, s, strlen(s));
}

void trig_set_qrg(struct trig *trig, double qrg){
    char s[40];

    if (!trig) return;
    g_snprintf(s, sizeof(s), "FREQ;%f;\n", qrg - get_rig_lo(aband));
    write(trig->pipe_write, s, strlen(s));
}

void trig_resend_freq(struct trig *trig){
    char s[40];

    if (!trig) return;
    g_snprintf(s, sizeof(s), "RESENDFREQ;\n");
    write(trig->pipe_write, s, strlen(s));
}

int trig_main(rig_model_t model, char *filename){
    char s[256];
    int ret;
    RIG *rig;
    freq_t freq, oldfreq;
    

    dbg("trig_main\n");
    signal(SIGINT, SIG_DFL);
    signal(SIGSEGV, SIG_DFL);

    freq = oldfreq = (freq_t)0;

    rig = rig_init(model);
    if (!rig){
        printf("TRIG;!;Can't init rig %d\n", (int)model);
        write(tpipe->threadpipe_write, s, strlen(s));
        return 1;
    }


    
    safe_strncpy(rig->state.rigport.pathname, filename, FILPATHLEN);
    ret = rig_open(rig);
    if (ret != RIG_OK){
        printf("TRIG;!;Can't open rig %d (%s)\n", (int)model, rig->state.rigport.pathname);
        rig_cleanup(rig);
        return 3;
    }
    
    if (fcntl(0, F_SETFL, O_NONBLOCK)){
        printf("TRIG;e;Can't set O_NONBLOCK\n");
        rig_close(rig);
        rig_cleanup(rig);
        return 4;
    }
    
    while(1){
        int status;
        struct zstring *zs;
        char *cmd;
        
        s[sizeof(s)-1]='\0';
        if (fgets(s, sizeof(s)-1, stdin) != NULL){
            //fprintf(stderr, "s=%s\n", s);
            zs = zstrdup(s);
            cmd = ztokenize(zs, 1);

            if (strcasecmp(cmd, "MODE")==0){
                rmode_t mode;

                mode = (rmode_t)atoi(ztokenize(zs, 0));
                status = rig_set_mode(rig, RIG_VFO_CURR, mode, RIG_PASSBAND_NORMAL);
                fprintf(stderr, "mode=%d status=%d\n", mode, status);
            }

            if (strcasecmp(cmd, "FREQ")==0){
                freq_t fr;
                
                sscanf(ztokenize(zs, 0), "%"SCNfreq, &fr);
                status = rig_set_freq(rig, RIG_VFO_CURR, fr);
                fprintf(stderr, "freq=%"PRIfreq"\n", fr);
            }

            if (strcasecmp(cmd, "RESENDFREQ")==0){
                printf("TRIG;qrg;%"PRIfreq"\n", freq);
            }
            zfree(zs);
        }
        
        status = rig_get_freq(rig, RIG_VFO_CURR, &freq);
        if (status == RIG_OK){  
            //dbg("TRIG;qrg;%"PRIfreq"\n", freq);
            if (freq != oldfreq){
                oldfreq = freq;
                printf("TRIG;qrg;%"PRIfreq"\n", freq);
            }
        }  
    }
    return 0;
}

int trig_save_riglist(const struct rig_caps *caps, void *data){
    GIndexArray *riglist = (GIndexArray *)data;
    g_index_array_add(riglist, (struct rig_caps *)caps);
    return -1;
}

int trig_compare (const void *a, const void *b){
    struct rig_caps **rca, **rcb;
    int d;

    rca = (struct rig_caps **)a;
    rcb = (struct rig_caps **)b;

    if (!*rca && !*rcb) return 0;
    if (!*rca) return -1;
    if (!*rcb) return +1;

    d = strcasecmp((*rca)->mfg_name, (*rcb)->mfg_name);
    if (d) return d;
    return strcasecmp((*rca)->model_name, (*rcb)->model_name);
}


#endif
